// \[geXg ̂Q
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <algorithm>
#include <functional>
#include <vector>
#include <string>


template<typename BidiIte> inline
void insert_sort(const BidiIte& first, const BidiIte& last)
{
	typedef typename std::iterator_traits<BidiIte>::value_type	T;
	using std::iter_swap;

	BidiIte i = first;
	while (++i != last) {
		T		tmp( *i );
		BidiIte j = i;
		BidiIte k = i;
		--k;
		while (j != first && tmp < *k) {
			*j = *k;
			--j;
			--k;
		}
		*j = tmp;
	}
}


template<typename BidiIte, class C> inline
void insert_sort(const BidiIte& first, const BidiIte& last, C cmp)
{
	typedef typename std::iterator_traits<BidiIte>::value_type	T;
	using std::iter_swap;

	BidiIte i = first;
	while (++i != last) {
		T		tmp( *i );
		BidiIte j = i;
		BidiIte k = i;
		--k;
		while (j != first && cmp(tmp, *k)) {
			*j = *k;
			--j;
			--k;
		}
		*j = tmp;
	}
}


template<typename RAIte>
void quick_sort(RAIte first, RAIte last)
{
	typedef typename std::iterator_traits<RAIte>::value_type	T;
	using std::iter_swap;

	std::size_t		size = last - first;
	if (size <= 32) {
		if (size > 1)
			insert_sort(first, last);
		return;
	}
	const T m( *( first + size / 2 ) );
	RAIte 	l = first;
	RAIte 	r = last;
	for (;;) {
		while (*l < m)
			++l;
		do {
			--r;
		} while (m < *r);
		if (!(l < r))
			break;
		iter_swap(l, r);
		++l;
	}
	quick_sort(first, l   );
	quick_sort(l    , last);
}



template<typename RAIte, class C>
void quick_sort(RAIte first, RAIte last, C cmp)
{
	typedef typename std::iterator_traits<RAIte>::value_type	T;
	using std::iter_swap;

	std::size_t		size = last - first;
	if (size <= 32) {
		if (size > 1)
			insert_sort(first, last, cmp);
		return;
	}
	const T m( *( first + size / 2 ) );
	RAIte 	l = first;
	RAIte 	r = last;
	for (;;) {
		while (cmp(*l, m))
			++l;
		do {
			--r;
		} while (cmp(m, *r));
		if (!(l < r))
			break;
		iter_swap(l, r);
		++l;
	}
	quick_sort(first, l   , cmp);
	quick_sort(l    , last, cmp);
}



template<typename BidiIte>
void bidi_quick_sort(BidiIte first, BidiIte last)
{
	typedef typename std::iterator_traits<BidiIte>::value_type	T;

	using std::iter_swap;
	std::size_t	size = std::size_t(std::distance(first, last));
	if (size <= 32) {
		if (size > 1)
			insert_sort(first, last);
		return;
	}
	BidiIte	ite( first );
	std::advance( ite, size / 2 );
	const T m( *ite );
	BidiIte 	l = first;
	BidiIte 	r = last;
	std::size_t lc = 0;
	std::size_t rc = size;
	for (;;) {
		while (*l < m)
			++l, ++lc;
		do {
			--r, --rc;
		} while (m < *r);
		if (lc >= rc)
			break;
		iter_swap(l, r);
		++l, ++lc;
	}
	bidi_quick_sort(first, l   );
	bidi_quick_sort(l    , last);
}



template<typename BidiIte, class C>
void bidi_quick_sort(BidiIte first, BidiIte last, C cmp)
{
	typedef typename std::iterator_traits<BidiIte>::value_type	T;

	using std::iter_swap;
	std::size_t	size = std::size_t(std::distance(first, last));
	if (size <= 32) {
		if (size > 1)
			insert_sort(first, last, cmp);
		return;
	}
	BidiIte	ite( first );
	std::advance( ite, size / 2 );
	const T m( *ite );
	BidiIte 	l = first;
	BidiIte 	r = last;
	std::size_t lc = 0;
	std::size_t rc = size;
	for (;;) {
		while (cmp(*l, m))
			++l, ++lc;
		do {
			--r, --rc;
		} while (cmp(m, *r));
		if (lc >= rc)
			break;
		iter_swap(l, r);
		++l, ++lc;
	}
	bidi_quick_sort(first, l   , cmp);
	bidi_quick_sort(l    , last, cmp);
}



// ===========================================================================
#ifdef _WIN32
#include <windows.h>
typedef unsigned __int64 PerfCnt_tick_t;
PerfCnt_tick_t PerfCnt_getTick()    { PerfCnt_tick_t c; QueryPerformanceCounter((LARGE_INTEGER*)&c); return c; }
PerfCnt_tick_t PerfCnt_tickPerSec() { static PerfCnt_tick_t s = 0; if (!s) QueryPerformanceFrequency((LARGE_INTEGER*)&s); return s; }
#else
#include <time.h>
typedef clock_t PerfCnt_tick_t;
#define PerfCnt_getTick() 			clock()
#define PerfCnt_tickPerSec() 		CLOCKS_PER_SEC
#endif


// ===========================================================================

template<typename T>
T random() {
	return T(rand());
}

template<>
std::string random<std::string>() {
	char buf[128];
	sprintf(buf, "%d%d%d", rand(),rand(),rand());
	return std::string(buf);
}


#define BENCHMARK(name,SORT,T,cmpFlg,size,count,seed) do {			\
	std::vector<T> data;											\
	srand(seed);													\
    for (unsigned i = 0; i < size * count; ++i) {					\
        data.push_back(random<T>());								\
    }																\
    unsigned 				ofs        = 0;							\
    PerfCnt_tick_t	start_time = PerfCnt_getTick();					\
	for (unsigned j = 0; j < count; ++j) {							\
		if (cmpFlg == 0)											\
	    	SORT(&data[ofs], &data[ofs]+size);						\
	    else														\
	    	SORT(&data[ofs], &data[ofs]+size, std::less<T>() );		\
	    ofs        += size;											\
	}																\
	PerfCnt_tick_t	elapsed    = PerfCnt_getTick() - start_time;	\
    printf("%s:%13.3fʕb (%eb)\n", name							\
    		, elapsed * 1000000.0 / (PerfCnt_tickPerSec() * count)	\
    		, elapsed / double(PerfCnt_tickPerSec() * count) );		\
} while (0)


#ifndef TYPE
#define TYPE	int
#endif

//
int main(int argc, char **argv) {
    size_t  size  = (argc >= 2) ? atoi(argv[1]) : 100;
    size_t  count = (argc >= 3) ? atoi(argv[2]) : 1;
    int    seed  = 1; //time(0);
    printf(" %d,  %d\n", size, count);
	BENCHMARK("std::sort(arg2)      ", std::sort,       TYPE, 0, size, count, seed);
	BENCHMARK("std::sort(arg3)      ", std::sort,       TYPE, 1, size, count, seed);
	BENCHMARK("quick sort(arg2)     ", quick_sort,      TYPE, 0, size, count, seed);
	BENCHMARK("quick sort(arg3)     ", quick_sort,      TYPE, 1, size, count, seed);
	BENCHMARK("quick sort:bidi(arg2)", bidi_quick_sort, TYPE, 0, size, count, seed);
	BENCHMARK("quick sort:bidi(arg3)", bidi_quick_sort, TYPE, 1, size, count, seed);
    return 0;
}
