// \[geXg ̂R : list f[^̃\[g.
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <algorithm>
#include <functional>
#include <vector>
#include <list>
#include <string>


template<typename BidiIte, class C >
void bidi_quick_sort(BidiIte first, BidiIte last, C cmp )
{
	typedef typename std::iterator_traits<BidiIte>::value_type	T;

	using std::iter_swap;
	std::size_t	size = std::size_t(std::distance(first, last));
	if (size <= 32) {
		if (size <= 1)
			return;
		BidiIte i = first;
		while (++i != last) {
			T		tmp( *i );
			BidiIte j = i;
			BidiIte k = i;
			do {
				--k;
				if (!cmp(tmp, *k))
					break;
				*j = *k;
				--j;
			} while (k != first);
			*j = tmp;
		}
		return;
	}
	BidiIte	ite( first );
	std::advance( ite, size / 2 );
	const T m( *ite );
	BidiIte 	l = first;
	BidiIte 	r = last;
	std::size_t lc = 0;
	std::size_t rc = size;
	for (;;) {
		while (cmp(*l, m))
			++l, ++lc;
		do {
			--r, --rc;
		} while (cmp(m, *r));
		if (lc >= rc)
			break;
		iter_swap(l, r);
		++l, ++lc;
	}
	bidi_quick_sort(first, l   , cmp);
	bidi_quick_sort(l    , last, cmp);
}



template<typename BidiIte> inline
void bidi_quick_sort(BidiIte first, BidiIte last) {
	bidi_quick_sort( first, last, std::less<typename std::iterator_traits<BidiIte>::value_type>() );
}



// ===========================================================================
#ifdef _WIN32
#include <windows.h>
typedef unsigned __int64 PerfCnt_tick_t;
PerfCnt_tick_t PerfCnt_getTick()    { PerfCnt_tick_t c; QueryPerformanceCounter((LARGE_INTEGER*)&c); return c; }
PerfCnt_tick_t PerfCnt_tickPerSec() { static PerfCnt_tick_t s = 0; if (!s) QueryPerformanceFrequency((LARGE_INTEGER*)&s); return s; }
#else
#include <time.h>
typedef clock_t PerfCnt_tick_t;
#define PerfCnt_getTick() 			clock()
#define PerfCnt_tickPerSec() 		CLOCKS_PER_SEC
#endif



// ===========================================================================

template<typename T>
T random() {
	return T(rand());
}

template<>
std::string random<std::string>() {
	char buf[128];
	sprintf(buf, "%d%d%d", rand(),rand(),rand());
	return std::string(buf);
}



// ===========================================================================
#ifdef USE_SMPCLASS
#if USE_SMPCLASS != 1
#include <tr1/array>
#endif

class SmpClass {
public:
	~SmpClass() { free(name_); }
	SmpClass() : name_(0), x_(0),y_(0),z_(0),argb_(0xFFFFFFFF),mode_(0) { /*buf_.resize(256);*/ }

	SmpClass(const SmpClass& r)
		: name_(strdup(r.name_)), x_(r.x_), y_(r.y_), z_(r.z_), argb_(r.argb_), mode_(r.mode_), buf_(r.buf_) {}

	SmpClass(const char* name, double x, double y, double z, unsigned argb=0xffffffff, unsigned mode=0)
		: name_(strdup(name)), x_(x), y_(y), z_(z), argb_(argb), mode_(mode)
	{
	  #if USE_SMPCLASS == 1
		buf_.resize(256);
	  #endif
	}

	SmpClass& operator=(const SmpClass& r) { SmpClass(r).swap( *this ); return *this; }

	void swap(SmpClass& r);
	bool operator<(const SmpClass& r) const;

private:
	char*						name_;
	double						x_, y_, z_;
	unsigned					argb_;
	unsigned					mode_;
  #if USE_SMPCLASS == 1
	std::vector<char>			buf_;
  #else
	std::tr1::array<char,160>	buf_;
  #endif
};


inline void SmpClass::swap(SmpClass& r) {
	using std::swap;
	swap(name_, r.name_);
	swap(x_   , r.x_   );
	swap(y_   , r.y_   );
	swap(z_   , r.z_   );
	swap(argb_, r.argb_);
	swap(mode_, r.mode_);
	swap(buf_ , r.buf_ );
}


inline bool SmpClass::operator<(const SmpClass& r) const {
	if (z_ < r.z_)
		return 1;
	if (z_ > r.z_)
		return 0;
	if (y_ < r.y_)
		return 1;
	if (y_ > r.y_)
		return 0;
	return strcmp(name_, r.name_) < 0;
}


template<>
SmpClass random<SmpClass>() {
	char name[128];
	sprintf(name, "%c%c%d", 'A'+rand()%26,'a'+rand()%26,'a'+rand()%26, rand());
	return SmpClass(name, rand(), rand(), rand());
}


#define TYPE		SmpClass
#endif



// ===========================================================================

template<typename T>
double benchmark1(unsigned size, unsigned count, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::vector<T> data;
		data.reserve(size);
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());
    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
   		bidi_quick_sort(data.begin(), data.end() );
		elapsed    += PerfCnt_getTick() - start_time;
	}
	return elapsed / double(PerfCnt_tickPerSec() * count);
}


template<typename T>
double benchmark2(unsigned size, unsigned count, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::list<T> data;
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());
    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
   		bidi_quick_sort( data.begin(), data.end() );
		elapsed    += PerfCnt_getTick() - start_time;
	}
	return elapsed / double(PerfCnt_tickPerSec() * count);
}


template<typename T>
double benchmark3(unsigned size, unsigned count, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::list<T> data;
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());
    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
   		data.sort();
		elapsed    += PerfCnt_getTick() - start_time;
	}
	return elapsed / double(PerfCnt_tickPerSec() * count);
}


template<typename T>
struct ptr_less {
	bool operator()(T l, T r) const {
		return *l < *r;
	}
};


template<typename T>
double benchmark4(unsigned size, unsigned count, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::list<T> data;
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());

    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
	    std::vector<T*>	ent(size);
	    typename std::vector<T*>::iterator vi = ent.begin();
	    for (typename std::list<T>::iterator li = data.begin(); li != data.end(); ++li) {
			*vi = &*li;
			++vi;
		}
   		bidi_quick_sort(ent.begin(), ent.end(), ptr_less<T*>() );
		elapsed    += PerfCnt_getTick() - start_time;
	}
	return elapsed / double(PerfCnt_tickPerSec() * count);
}




#ifndef TYPE
#define TYPE	int
#endif

struct St {
	size_t		size;
	size_t		count;
	double		elaped[4];
};


//
int main(int argc, char **argv) {
	if (argc < 2) {
		printf("usage> sort_test3 size1 count1 [size2 count2 ...]\n");
		return 1;
	}
	std::vector<St> st;
	unsigned n = 0;
	for (unsigned i = 1; i < argc; i += 2) {
		St	t;
		t.size   = atoi(argv[i]);
    	t.count  = (i+1 < argc) ? atoi(argv[i+1]) : 1;
    	std::fill(t.elaped, t.elaped+4, 0.0);
    	st.push_back(t);
    	++n;
    }

    printf("\nsort_test3 : sizeof(TYPE)=%dbytes\n", sizeof(TYPE));
    int    seed  = 1; //time(0);
	for (unsigned i = 0; i < n; ++i) {
		St& t = st[i];
	    printf(" %9d,  %9d\n", t.size, t.count);
		t.elaped[0] = benchmark1<TYPE>(t.size, t.count, seed);
		t.elaped[1] = benchmark2<TYPE>(t.size, t.count, seed);
		t.elaped[2] = benchmark3<TYPE>(t.size, t.count, seed);
		t.elaped[3] = benchmark4<TYPE>(t.size, t.count, seed);
	}

	// ʕ\.
	static const char* ttl[] = {
		"vector quick sort               ",
		"list   quick sort               ",
		"std::list.sort                    ",
		"listvfւ̃|C^vectorsort",
		NULL,
	};
	printf("\n");
	printf("%34s ,", "");
	for (unsigned i = 0; i < n; ++i) {
		printf("%9d,", st[i].size);
	};
	printf("\n");
	for (unsigned j = 0; ttl[j]; ++j) {
		printf("%-34s ,", ttl[j]);
		for (unsigned i = 0; i < n; ++i) {
			printf("%11.3f,", st[i].elaped[j]*1000000.0);
		};
		printf("\n");
	}
    return 0;
}
