// \[geXg ̂S X}[g|C^pꍇ.
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <algorithm>
#include <functional>
#include <vector>
#include <list>
#include <string>

#if _MSC_VER >= 1500
#include <memory>
using namespace std::tr1;
#else
#include <boost/shared_ptr.hpp>
using namespace boost;
#endif


template<typename RAIte, class C >
void quick_sort(RAIte first, RAIte last, C cmp)
{
	typedef typename std::iterator_traits<RAIte>::value_type	T;

	using std::iter_swap;
	std::size_t	size = last - first;
	if (size <= 32) {
		if (size <= 1)
			return;
		RAIte i = first;
		while (++i != last) {
			T		tmp( *i );
			RAIte j = i;
			RAIte k = i;
			do {
				--k;
				if (!cmp(tmp, *k))
					break;
				*j = *k;
				--j;
			} while (k != first);
			*j = tmp;
		}
		return;
	}
	const T m( *(first + size/2) );
	RAIte 	l = first;
	RAIte 	r = last;
	for (;;) {
		while (cmp(*l, m))
			++l;
		do {
			--r;
		} while (cmp(m, *r));
		if (!(l < r))
			break;
		iter_swap(l, r);
		++l;
	}
	quick_sort(first, l   , cmp);
	quick_sort(l    , last, cmp);
}



template<typename RAIte> inline
void quick_sort(RAIte first, RAIte last) {
	quick_sort( first, last, std::less<typename std::iterator_traits<RAIte>::value_type>() );
}



// ===========================================================================
#ifdef _WIN32
#include <windows.h>
typedef unsigned __int64 PerfCnt_tick_t;
PerfCnt_tick_t PerfCnt_getTick()    { PerfCnt_tick_t c; QueryPerformanceCounter((LARGE_INTEGER*)&c); return c; }
PerfCnt_tick_t PerfCnt_tickPerSec() { static PerfCnt_tick_t s = 0; if (!s) QueryPerformanceFrequency((LARGE_INTEGER*)&s); return s; }
#else
#include <time.h>
typedef clock_t PerfCnt_tick_t;
#define PerfCnt_getTick() 			clock()
#define PerfCnt_tickPerSec() 		CLOCKS_PER_SEC
#endif



// ===========================================================================

template<typename T>
T random() {
	return T(rand());
}

template<>
std::string random<std::string>() {
	char buf[128];
	sprintf(buf, "%d%d%d", rand(),rand(),rand());
	return std::string(buf);
}



// ===========================================================================
#ifdef USE_SMPCLASS
#if USE_SMPCLASS != 1
#include <tr1/array>
#endif
#include <boost/intrusive_ptr.hpp>

class SmpClass {
public:
	~SmpClass() { free(name_); }
	SmpClass() : name_(0), x_(0),y_(0),z_(0),argb_(0xFFFFFFFF),refCnt_(0) { /*buf_.resize(256);*/ }

	SmpClass(const SmpClass& r)
		: name_(strdup(r.name_)), x_(r.x_), y_(r.y_), z_(r.z_), argb_(r.argb_), refCnt_(0), buf_(r.buf_) {}

	SmpClass(const char* name, double x, double y, double z, unsigned argb=0xffffffff)
		: name_(strdup(name)), x_(x), y_(y), z_(z), argb_(argb), refCnt_(0)
	{
	  #if USE_SMPCLASS == 1
		buf_.resize(256);
	  #endif
	}

	SmpClass& operator=(const SmpClass& r) { SmpClass(r).swap( *this ); return *this; }

	void swap(SmpClass& r);
	bool operator<(const SmpClass& r) const;

	void     incRef() { if (this) ++refCnt_; }
	void     decRef_and_release() { if (this && refCnt_) { if (!(--refCnt_)) delete this; } }
	unsigned ref_count() const { return refCnt_; }

private:
	char*						name_;
	double						x_, y_, z_;
	unsigned					argb_;
	unsigned					refCnt_;
  #if USE_SMPCLASS == 1
	std::vector<char>			buf_;
  #else
	std::tr1::array<char,160>	buf_;
  #endif
};


inline bool SmpClass::operator<(const SmpClass& r) const {
	if (z_ < r.z_)
		return 1;
	if (z_ > r.z_)
		return 0;
	if (y_ < r.y_)
		return 1;
	if (y_ > r.y_)
		return 0;
	return strcmp(name_, r.name_) < 0;
}


inline void SmpClass::swap(SmpClass& r) {
	using std::swap;
	swap(name_, r.name_);
	swap(x_   , r.x_   );
	swap(y_   , r.y_   );
	swap(z_   , r.z_   );
	swap(argb_, r.argb_);
	swap(refCnt_, r.refCnt_);
	swap(buf_ , r.buf_ );
}


void intrusive_ptr_add_ref(SmpClass* p) { p->incRef(); }
void intrusive_ptr_release(SmpClass* p) { p->decRef_and_release(); }



template<>
SmpClass random<SmpClass>() {
	char name[128];
	sprintf(name, "%c%c%d", 'A'+rand()%26,'a'+rand()%26,'a'+rand()%26, rand());
	return SmpClass(name, rand(), rand(), rand());
}



#define TYPE		SmpClass
#endif



// ===========================================================================

template<typename T>
struct ptr_less {
	bool operator()(T l, T r) const {
		return *l < *r;
	}
};



// vector<T>Ɋmۂf[^ ̗vfւ̃|C^̔ză\[g.
template<typename T>
double benchmark1(unsigned size, unsigned count=1, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::vector<T> data;
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());

    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
	    std::vector<T*>	ent(size);
	    typename std::vector<T*>::iterator vi = ent.begin();
	    for (typename std::vector<T>::iterator li = data.begin(); li != data.end(); ++li) {
			*vi = &*li;
			++vi;
		}
   		quick_sort(ent.begin(), ent.end(), ptr_less<T*>() );
		elapsed    += PerfCnt_getTick() - start_time;

		for (unsigned i = 0; i < size-1; ++i) {	// \[g `FbN.
			if (*ent[i+1] < *ent[i])
				printf("error %d : > \n", i);
		}
	}
	return elapsed * 1.0 / (PerfCnt_tickPerSec() * count);
	//printf("%s:%13.3fʕb\n", name, elapsed * 1000000.0 / (PerfCnt_tickPerSec() * count));
}


// vector \[g
template<typename T>
double benchmark2(unsigned size, unsigned count=1, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::vector<T> data;
		data.reserve(size);
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());
    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
   		quick_sort(data.begin(), data.end() );
		elapsed    += PerfCnt_getTick() - start_time;

		for (unsigned i = 0; i < size-1; ++i) {	// \[g `FbN.
			if (data[i+1] < data[i])
				printf("error %d : > \n", i);
		}
	}
	return elapsed * 1.0 / (PerfCnt_tickPerSec() * count);
}



// vector<T>Ɋmۂf[^ ̗vfւ̃|C^̔ză\[gAvector<T>f[^č\z.
template<typename T>
double benchmark3(unsigned size, unsigned count=1, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		std::vector<T> data;
		data.reserve( size );
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back(random<T>());

    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
		{
		    std::vector<T*>	ent(size);
		    typename std::vector<T*>::iterator vi = ent.begin();
		    for (typename std::vector<T>::iterator it = data.begin(); it != data.end(); ++it) {
				*vi = &*it;
				++vi;
			}
	   		quick_sort(ent.begin(), ent.end(), ptr_less<T*>() );
			std::vector<T> data2;
			data2.reserve( size );
			for (unsigned i = 0; i < size; ++i)
				data2.push_back( *ent[i] );
			std::swap(data, data2);
		}
		elapsed    += PerfCnt_getTick() - start_time;

		for (unsigned i = 0; i < size-1; ++i) {	// \[g `FbN.
			if (data[i+1] < data[i])
				printf("error %d : > \n", i);
		}
	}
	return elapsed * 1.0 / (PerfCnt_tickPerSec() * count);
}



// X}[g|C^vector\[g.
template<typename T>
double benchmark4(unsigned size, unsigned count=1, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		typedef shared_ptr<T>  T_ptr;
		std::vector< T_ptr >   data;
		data.reserve( size );
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back( T_ptr( new T( random<T>() ) )  );

    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
   		quick_sort(data.begin(), data.end(), ptr_less<T_ptr>() );
		elapsed    += PerfCnt_getTick() - start_time;

		for (unsigned i = 0; i < size-1; ++i) {	// \[g `FbN.
			if (*data[i+1] < *data[i])
				printf("error %d : > \n", i);
		}
	}
	return elapsed * 1.0 / (PerfCnt_tickPerSec() * count);
}


#ifdef USE_SMPCLASS
// intrusive_ptrvector\[g.
template<typename T>
double benchmark5(unsigned size, unsigned count=1, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		typedef boost::intrusive_ptr<T>  T_ptr;
		std::vector< T_ptr > data;
		data.reserve( size );
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back( T_ptr( new T( random<T>() ) )  );

    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
   		quick_sort(data.begin(), data.end(), ptr_less<T_ptr>() );
		elapsed    += PerfCnt_getTick() - start_time;

		for (unsigned i = 0; i < size-1; ++i) {	// \[g `FbN.
			if (*data[i+1] < *data[i])
				printf("error %d : > \n", i);
		}
	}
	return elapsed * 1.0 / (PerfCnt_tickPerSec() * count);
}


// vector<T>Ɋmۂf[^ ̗vfւ̃|C^̔ză\[gAvector<T>f[^č\z.
template<typename T>
double benchmark6(unsigned size, unsigned count=1, int seed=1) {
	srand(seed);
	PerfCnt_tick_t	elapsed = 0;
	for (unsigned j = 0; j < count; ++j) {
		typedef boost::intrusive_ptr<T>  T_ptr;
		std::vector< T_ptr > data;
		data.reserve( size );
	    for (unsigned i = 0; i < size; ++i)
	        data.push_back( T_ptr( new T( random<T>() ) )  );

    	PerfCnt_tick_t	start_time = PerfCnt_getTick();
		{
		    std::vector<T*>	ent(size);
		    typename std::vector<T*>::iterator vi = ent.begin();
		    for (typename std::vector<T_ptr>::iterator it = data.begin(); it != data.end(); ++it) {
				*vi = it->get();
				++vi;
			}
	   		quick_sort(ent.begin(), ent.end(), ptr_less<T*>() );
			{
				std::vector<T_ptr> data2;
				data2.reserve( size );
				for (unsigned i = 0; i < size; ++i)
					data2.push_back( T_ptr(ent[i]) );
				std::swap(data, data2);
			}
		}
		elapsed    += PerfCnt_getTick() - start_time;

		for (unsigned i = 0; i < size; ++i) {	// QƐĂȂ`FbN.
			if (data[i]->ref_count() != 1)
				printf("error %d\n", i);
		}
		for (unsigned i = 0; i < size-1; ++i) {	// \[g `FbN.
			if (*data[i+1] < *data[i])
				printf("error %d : > \n", i);
		}
	}
	return elapsed * 1.0 / (PerfCnt_tickPerSec() * count);
}
#endif



#ifndef TYPE
#define TYPE	int
#endif

struct St {
	size_t		size;
	size_t		count;
	double		elaped[6];
};

//
int main(int argc, char **argv) {
	if (argc < 2) {
		printf("usage> sort_test4 size1 count1 [size2 count2 ...]\n");
		return 1;
	}
	std::vector<St> st;
	unsigned n = 0;
	for (unsigned i = 1; i < argc; i += 2) {
		St	t;
		t.size   = atoi(argv[i]);
    	t.count  = (i+1 < argc) ? atoi(argv[i+1]) : 1;
    	std::fill(t.elaped, t.elaped+6, 0.0);
    	st.push_back(t);
    	++n;
    }
    printf("\nsort_test4 : sizeof(TYPE)=%dbytes\n", sizeof(TYPE));
    int    seed  = 1; //time(0);
	for (unsigned i = 0; i < n; ++i) {
		St& t = st[i];
	    printf(" %9d,  %9d\n", t.size, t.count);
		t.elaped[0] = benchmark1<TYPE>(t.size, t.count, seed);	// "vectorvfւ̃|C^vector݂̂sort       "
		t.elaped[1] = benchmark2<TYPE>(t.size, t.count, seed);	// "vector sort                                  "
		t.elaped[2] = benchmark3<TYPE>(t.size, t.count, seed);	// "|C^Ń\[gāǍʂvectorč\z   "
		t.elaped[3] = benchmark4<TYPE>(t.size, t.count, seed);	// "shared_ptrvector sort                      "
	  #ifdef USE_SMPCLASS
		t.elaped[4] = benchmark5<TYPE>(t.size, t.count, seed); // "intrusive_ptrvector sort                   "
		t.elaped[5] = benchmark6<TYPE>(t.size, t.count, seed); // "intrusive_ptrvector𐶃|C^\[gōč\z"
	  #endif
	}

	// ʕ\.
	static const char* ttl[] = {
		"vectorvfւ̃|C^vector݂̂sort       ",	// 1
		"vector sort                                  ",	// 2
		"|C^Ń\[gāǍʂvectorč\z   ",	// 3
		"shared_ptrvector sort                      ",	// 4
	  #ifdef USE_SMPCLASS
		"intrusive_ptrvector sort                   ",	// 5
		"intrusive_ptrvector𐶃|C^\[gōč\z",	// 6
	  #endif
		NULL,
	};
	printf("\n");
	printf("%47s ,", "");
	for (unsigned i = 0; i < n; ++i) {
		printf("%9d,", st[i].size);
	};
	printf("\n");
	for (unsigned j = 0; ttl[j]; ++j) {
		printf("%-47s ,", ttl[j]);
		for (unsigned i = 0; i < n; ++i) {
			printf("%11.3f,", st[i].elaped[j]*1000000.0);
		};
		printf("\n");
	}
    return 0;
}
